"""Chip8 interpreter."""

import array
from collections import namedtuple
import pathlib
import time

from .clock import stabilize_frame
from .code import dispatch
from .config import Config
from .debug import Disassembler
from .errors import ChippyError
from .processor import ExecutionUnit
from .status import Mode
from .window import buzz, Window

class Chippy:
    def __init__(self, config=Config()):
        """Initialize RAM, registers, stack, IO and sprite data."""
        self.ram = bytearray([0x00] * 4096)

        self.registers = bytearray([0x00] * 16)
        self.I = 0x0000
        self.sound_timer = 0x00
        self.delay_timer = 0x00
        self.program_counter = 0x0200
        self.stack_pointer = 0x00
        self.stack = array.array('H', [0x0000] * 16)

        self.keypad = 0x0000
        self.display = None
        # 64-by-32 display

        self.initialize_display()
        self.initialize_sprite_data()

        self.status = Mode.STOP
        self.waiting = []

        self.config = config
        self.disassembler = Disassembler()
        self.execution_unit = ExecutionUnit(self)

    def initialize_display(self):
        """Clear display."""
        self.display = array.array('Q', [0x0000000000000000] * 32)

    def initialize_sprite_data(self):
        """Initialize sprite data in locates 0x000 to 0x050."""
        self.ram[:5]    = (0xf0, 0x90, 0x90, 0x90, 0xf0)
        self.ram[5:10]  = (0x20, 0x60, 0x20, 0x20, 0x70)
        self.ram[10:15] = (0Xf0, 0x10, 0xf0, 0x80, 0xf0)
        self.ram[15:20] = (0xf0, 0x10, 0xf0, 0x10, 0xf0)
        self.ram[20:25] = (0x90, 0x90, 0xf0, 0x10, 0x10)
        self.ram[25:30] = (0xf0, 0x80, 0xf0, 0x10, 0xf0)
        self.ram[30:35] = (0xf0, 0x80, 0xf0, 0x90, 0xf0)
        self.ram[35:40] = (0xf0, 0x10, 0x20, 0x40, 0x40)
        self.ram[40:45] = (0xf0, 0x90, 0xf0, 0x90, 0xf0)
        self.ram[45:50] = (0xf0, 0x90, 0xf0, 0x10, 0xf0)
        self.ram[50:55] = (0xf0, 0x90, 0xf0, 0x90, 0x90)
        self.ram[55:60] = (0xe0, 0x90, 0xe0, 0x90, 0xe0)
        self.ram[60:65] = (0xf0, 0x80, 0x80, 0x80, 0xf0)
        self.ram[65:70] = (0xe0, 0x90, 0x90, 0x90, 0xe0)
        self.ram[70:75] = (0xf0, 0x80, 0xf0, 0x80, 0xf0)
        self.ram[75:80] = (0xf0, 0x80, 0xf0, 0x80, 0x80)

    def jump(self, target):
        """Jump to target location."""
        if target < 0x200 or target >= len(self.ram):
            raise ChippyError(f"Invalid jump target: {target:#05x}")
        self.program_counter = target

    def load(self, program: pathlib.Path):
        """Load program into address 0x200."""
        binary = program.read_bytes()
        size = len(binary)
        if size >= len(self.ram) - 0x200:
            raise ChippyError("Ran out of memory.")
        self.ram[0x200:size + 0x200] = binary

    def fetch(self):
        """Fetch current instruction."""
        msb = self.ram[self.program_counter]
        lsb = self.ram[self.program_counter + 1]
        return (msb << 8) | lsb

    def increment(self):
        """Increment program counter.

        This is called by instruction handlers.
        """
        self.program_counter += 2
        self.program_counter &= 0x0fff

    def cycle(self):
        """Simulate one cycle."""
        if not self.waiting:
            instruction = self.fetch()
            self.increment()
            print(dispatch(instruction, self.disassembler))
            dispatch(instruction, self.execution_unit)

    def countdown(self):
        """Decrement timers and perform timer-related actions."""
        if self.delay_timer > 0:
            self.delay_timer -= 1
        if self.sound_timer > 0:
            self.sound_timer -= 1
            buzz()

    def run(self):
        """Run program stored in memory."""
        self.status = Mode.RUN
        window = Window(self)
        window.init_screen()

        stages = (self.cycle, window.handle_events, window.render)

        timer_60Hz = 0.01667
        while self.status != Mode.STOP:
            if self.status == Mode.RUN:
                elapsed = stabilize_frame(self.config.clock_period, *stages)
                timer_60Hz -= elapsed
                if timer_60Hz <= 0:
                    timer_60Hz = 0.01667
                    self.countdown()
            elif self.status == Mode.PAUSE:
                window.handle_events()
                window.render()



import random
import sys

from .errors import ChippyError

def classify(instruction):
    """Classify instruction.

    Return name of instruction handler and arguments.
    """
    if instruction in (0x00e0, 0x00ee):
        return f"op_{instruction:04x}",

    opcode = instruction >> 12

    if 0 <= opcode <= 2:
        return f"op_{opcode}nnn", instruction & 0x0fff
    if 3 <= opcode <= 4:
        x = instruction & 0x0f00
        kk = instruction & 0x00ff
        return f"op_{opcode}xkk", x >> 8, kk
    if opcode == 5:
        # if instruction & 0xf00f == 0x5000
        if instruction & 0x000f == 0:
            x = instruction & 0x0f00
            y = instruction & 0x00f0
            return "op_5xy0", x >> 8, y >> 4
    if 6 <= opcode <= 7:
        x = instruction & 0x0f00
        kk = instruction & 0x00ff
        return f"op_{opcode}xkk", x >> 8, kk
    if opcode == 8:
        function = instruction & 0x000f
        x = instruction & 0x0f00
        y = instruction & 0x00f0
        if 0 <= function <= 7:
            return f"op_8xy{function}", x >> 8, y >> 4
        if function == 0xe:
            return f"op_8xye", x >> 8, y >> 4
    if opcode == 9:
        if instruction & 0x000f == 0:
            x = instruction & 0x0f00
            y = instruction & 0x00f0
            return "op_9xy0", x >> 8, y >> 4
    if 0xa <= opcode <= 0xb:
        return f"op_{opcode:1x}nnn", instruction & 0x0fff
    if opcode == 0xc:
        x = instruction & 0x0f00
        kk = instruction & 0x00ff
        return "op_cxkk", x >> 8, kk
    if opcode == 0xd:
        x = instruction & 0x0f00
        y = instruction & 0x00f0
        n = instruction & 0x000f
        return "op_dxyn", x >> 8, y >> 4, n
    if opcode == 0xe:
        function = instruction & 0x00ff
        x = instruction & 0x0f00
        if function == 0x9e:
            return "op_ex9e", x >> 8
        if function == 0xa1:
            return "op_exa1", x >> 8
    if opcode == 0xf:
        function = instruction & 0x00ff
        if function in (0x07, 0x0a, 0x15, 0x18, 0x1e, 0x29, 0x33, 0x55, 0x65):
            x = instruction & 0x0f00
            return f"op_fx{function:02x}", x >> 8
    return "",

def dispatch(instruction, impl):
    """Run instruction on implementation."""

    opcode = instruction >> 12
    x = (instruction & 0xf00) >> 8
    y = (instruction & 0xf0) >> 4
    nnn = instruction & 0xfff
    kk = instruction & 0xff
    funct = instruction & 0xf

    if opcode == 0:
        if kk == 0xe0:
            return impl.op_00e0()
        elif kk == 0xee:
            return impl.op_00ee()
    elif opcode == 1:
        return impl.op_1nnn(nnn)
    elif opcode == 2:
        return impl.op_2nnn(nnn)
    elif opcode == 3:
        return impl.op_3xkk(x, kk)
    elif opcode == 4:
        return impl.op_4xkk(x, kk)
    elif opcode == 5:
        if funct == 0:
            return impl.op_5xy0(x, y)
    elif opcode == 6:
        return impl.op_6xkk(x, kk)
    elif opcode == 7:
        return impl.op_7xkk(x, kk)
    elif opcode == 8:
        if 0 <= funct <= 7:
            return {
                0: lambda: impl.op_8xy0(x, y),
                1: lambda: impl.op_8xy1(x, y),
                2: lambda: impl.op_8xy2(x, y),
                3: lambda: impl.op_8xy3(x, y),
                4: lambda: impl.op_8xy4(x, y),
                5: lambda: impl.op_8xy5(x, y),
                6: lambda: impl.op_8xy6(x, y),
                7: lambda: impl.op_8xy7(x, y),
            }[funct]()
        elif funct == 0xe:
            return impl.op_8xye(x, y)
    elif opcode == 9:
        if funct == 0:
            return impl.op_9xy0(x, y)
    elif opcode == 0xa:
        return impl.op_annn(nnn)
    elif opcode == 0xb:
        return impl.op_bnnn(nnn)
    elif opcode == 0xc:
        return impl.op_cxkk(x, kk)
    elif opcode == 0xd:
        n = funct
        return impl.op_dxyn(x, y, n)
    elif opcode == 0xe:
        if kk == 0x9e:
            return impl.op_ex9e(x)
        elif kk == 0xa1:
            return impl.op_exa1(x)
    elif opcode == 0xf:
        case = {
            0x07: lambda: impl.op_fx07(x),
            0x0a: lambda: impl.op_fx0a(x),
            0x15: lambda: impl.op_fx15(x),
            0x18: lambda: impl.op_fx18(x),
            0x1e: lambda: impl.op_fx1e(x),
            0x29: lambda: impl.op_fx29(x),
            0x33: lambda: impl.op_fx33(x),
            0x55: lambda: impl.op_fx55(x),
            0x65: lambda: impl.op_fx65(x),
        }.get(kk)
        if case:
            return case()
    return ChippyError(f"Invalid instruction: {instruction:#06x}")



"""chippy subprograms."""
from pathlib import Path
import errno
import sys

from .chippy import Chippy
from .config import Config

def list_roms():
    """List avaiable ROMs."""
    roms = Path(__file__).parent.joinpath("roms")
    for rom in roms.glob('*'):
        print(rom.name)

def find_rom(program):
    """Find ROM in current directory or in ROMs directory."""
    rom = Path(program)
    if rom.is_file():
        return rom
    rom = Path(__file__).parent.joinpath("roms", program)
    if rom.is_file():
        return rom

def run(program, config=Config()):
    """Run chip-8 program."""
    rom = find_rom(program)
    if rom is None:
        print(f"Program '{program}' not found.", file=sys.stderr)
        sys.exit(errno.ENOENT)
    chippy = Chippy(config)
    chippy.load(rom)
    chippy.run()



"""Chip-8 debugger."""

class Disassembler:
    def op_0nnn(self, nnn):
        """Jump to routine at nnn."""
        return "nop"

    def op_00e0(self):
        """Clear chip8 display."""
        return "clear"

    def op_00ee(self):
        """Return from subroutine."""
        return "return"

    def op_1nnn(self, nnn):
        """Jump to location nnn."""
        return f"jump {nnn:#05x}"

    def op_2nnn(self, nnn):
        """Call subroutine at nnn."""
        return f"call {nnn:#05x}"

    def op_3xkk(self, x, kk):
        """Skip the next instruction if Vx == kk."""
        return f"skip if V{x:x} = {kk}"

    def op_4xkk(self, x, kk):
        """Skip next instruction if Vx != kk."""
        return f"skip if V{x:x} != {kk}"

    def op_5xy0(self, x, y):
        """Skip next instruction if Vx = Vy."""
        return f"skip if V{x:x} = V{y:x}"

    def op_6xkk(self, x, kk):
        """Put kk into Vx."""
        return f"V{x:x} = {kk}"

    def op_7xkk(self, x, kk):
        """Add kk to Vx."""
        return f"V{x:x} += {kk}"

    def op_8xy0(self, x, y):
        """Set Vx = Vy."""
        return f"V{x:x} = V{y:x}"

    def op_8xy1(self, x, y):
        """Set Vx = Vx OR Vy."""
        return f"V{x:x} |= V{y:x}"

    def op_8xy2(self, x, y):
        """Set Vx = Vx AND Vy."""
        return f"V{x:x} &= V{y:x}"

    def op_8xy3(self, x, y):
        """Set Vx = Vx XOR Vy."""
        return f"V{x:x} ^= V{y:x}"

    def op_8xy4(self, x, y):
        """Add Vy to Vx and set Vf to the carry bit."""
        return f"V{x:x} += V{y:x}, Vf = 'carry'"

    def op_8xy5(self, x, y):
        """Subtract Vy from Vx and set Vf to 1 if there's no borrow."""
        return f"V{x:x} -= V{y:x}, Vf = 'no borrow'"

    def op_8xy6(self, x, y):
        """Set Vx = Vy >> 1, and set Vf to the LSB prior to the shift.

        NOTE Descriptions differ between
        - http://mattmik.com/files/chip8/mastering/chip8.html
        - http://devernay.free.fr/hacks/chip8/C8TECH10.HTM
        """
        return f"Vf = lsb V{y:x}\n\tV{x:x} = V{y:x} >> 1"

    def op_8xy7(self, x, y):
        """Set Vx = Vy - Vx and set Vf to 1 if there's no borrow."""
        return f"V{x:x} = V{y:x} - V{x:x}, Vf = 'no borrow'"

    def op_8xye(self, x, y):
        """Set Vx = Vy << 1, and set Vf to the MSB prior to the shift.

        NOTE Descriptions differ between
        - http://mattmik.com/files/chip8/mastering/chip8.html
        - http://devernay.free.fr/hacks/chip8/C8TECH10.HTM
        """
        return f"Vf = msb V{y:x}\n\tV{x:x} = V{y:x} << 1"

    def op_9xy0(self, x, y):
        """Skip next instruction if Vx != Vy."""
        return f"skip if V{x:x} != V{y:x}"

    def op_annn(self, nnn):
        """Set I to nnn."""
        return f"I = {nnn:#05x}"

    def op_bnnn(self, nnn):
        """Jump to nnn + V0."""
        return f"jump V0 + {nnn:#05x}"

    def op_cxkk(self, x, kk):
        """Set Vx = random byte & kk."""
        return f"V{x:x} = random & {kk:#04x}"

    def op_dxyn(self, x, y, nibble):
        """Display n-byte sprite starting at memory location I at (Vx, Vy).

        The sprite is 8 pixels wide and n pixels tall.
        Set Vf = 1 iff any set pixels are unset.
        The sprite is drawn by XORing it with the display.
        """
        return f"display 8-by-{nibble} sprite at (V{x:x}, V{y:x})"

    def op_ex9e(self, x):
        """Skip next instruction if key with the value of Vx is pressed."""
        return f"skip if key @V{x:x} is pressed"

    def op_exa1(self, x):
        """Skip next instruction if key with the value of Vx is not pressed."""
        return f"skip if key @V{x:x} is not pressed"

    def op_fx07(self, x):
        """Set Vx to the delay timer value."""
        return f"V{x:x} = DT"

    def op_fx0a(self, x):
        """Wait for a key press and store the value of the key in Vx."""
        return f"V{x:x} = wait key"

    def op_fx15(self, x):
        """Set the delay timer to Vx."""
        return f"DT = V{x:x}"

    def op_fx18(self, x):
        """Set the sound timer to Vx."""
        return f"ST = V{x:x}"

    def op_fx1e(self, x):
        """Add Vx to I."""
        return f"I += V{x:x}"

    def op_fx29(self, x):
        """Set I to location of sprite for digit in Vx."""
        return f"I = sprite address of digit in V{x:x}"

    def op_fx33(self, x):
        """Store the BCD representation of Vx in memory locations I, I+1 and I+2."""
        return f"I[:2] = bcd of value in V{x:x}"

    def op_fx55(self, x):
        """Store registers V0 to Vx (inclusive) in memory starting at location I.

        The value of I gets incremented by x + 1 afterwards.
        """
        return f"I[:{x:x} + 1] = V[:{x:x} + 1]\n\tI += {x} + 1"

    def op_fx65(self, x):
        """Read registers V0 through Vx (inclusive) from memory starting at I.

        The value of I gets incremented by x + 1 afterwards.
        """
        return f"V[:{x:x} + 1] = I[:{x:x} + 1]\n\tI += {x} + 1"



"""Run chip8 interpreter."""

from argparse import ArgumentParser

from . import app
from .config import Config

if __name__ == "__main__":
    config = Config()

    parser = ArgumentParser(prog="Chippy", description="Run chip-8 emulator.")
    parser.add_argument("-l", "--list", action="store_true",
                        help="list available ROMs")
    parser.add_argument("-p", "--play", metavar="ROM", help="load chip-8 ROM")
    parser.add_argument("-c", "--colors", default=config.color_scheme,
                        help=f"color scheme (default={config.color_scheme!r})")
    parser.add_argument("-r", "--clock-rate", default=config.clock_rate, type=int,
                        help=f"clock rate in Hz (default={config.clock_rate!r})")
    args = parser.parse_args()

    config.color_scheme = args.colors
    config.clock_rate = args.clock_rate

    if args.list:
        app.list_roms()
    elif args.play:
        app.run(args.play, config)



from dataclasses import dataclass

@dataclass
class Config:
    color_off = (0, 0, 0)
    color_on = (255, 255, 255)
    clock_rate = 500

    @property
    def color_scheme(self):
        off = ",".join(str(a) for a in self.color_off)
        on = ",".join(str(a) for a in self.color_on)
        return f"{off}:{on}"

    @color_scheme.setter
    def color_scheme(self, scheme):
        off, on = scheme.split(':')
        self.color_off = tuple(int(a) for a in off.split(','))
        self.color_on = tuple(int(a) for a in on.split(','))

    @property
    def clock_period(self):
        return 1/self.clock_rate

    @clock_period.setter
    def clock_period(self, period):
        self.clock_rate = 1/period



"""Stabilize clock rate."""

import time

def stabilize_frame(period, *functions):
    """Sleep if the functions don't use up the entire frame period.

    Return elapsed time in functions.
    """
    start_time = time.time()
    for f in functions:
        f()
    elapsed = time.time() - start_time
    remaining = period - elapsed
    if remaining > 0:
        time.sleep(remaining)
    return elapsed



"""Chip8 display and keypad."""

from pathlib import Path

import pygame

from .status import Mode

def buzz():
    """Sound buzzer."""
    path = Path(__file__).with_name("data").joinpath("chime.wav")
    chime = pygame.mixer.Sound(str(path))
    pygame.mixer.Sound.play(chime)

def press(chip8, key):
    """Press key on chip-8 keypad."""
    mask = 1 << key
    chip8.keypad |= mask

    # Handle op_fx0a
    if chip8.waiting:
        index = chip8.waiting.pop()
        chip8.registers[index] = key

def release(chip8, key):
    """Release key on chip-8 keypad."""
    mask = 0xffff
    mask -= (1 << key)
    chip8.keypad &= mask

KEYS = {
    pygame.K_x: 0,
    pygame.K_1: 1,
    pygame.K_2: 2,
    pygame.K_3: 3,
    pygame.K_q: 4,
    pygame.K_w: 5,
    pygame.K_e: 6,
    pygame.K_a: 7,
    pygame.K_s: 8,
    pygame.K_d: 9,
    pygame.K_z: 0xa,
    pygame.K_c: 0xb,
    pygame.K_4: 0xc,
    pygame.K_r: 0xd,
    pygame.K_f: 0xe,
    pygame.K_v: 0xf,
}
def convert_key(key):
    """Convert pygame key constant to chip-8 key input."""
    return KEYS.get(key)

def draw_pixel(x, y, scale=1, color=(255, 255, 255)):
    """Draw pixel on the display."""
    surface = pygame.display.get_surface()
    rect = pygame.Rect(x * scale, y * scale, scale, scale)
    pygame.draw.rect(surface, color, rect)

class Window:
    def __init__(self, chip8):
        self.width = 64
        self.height = 32
        self.scale = 12

        self.chip8 = chip8
        self.color_off = self.chip8.config.color_off
        self.color_on = self.chip8.config.color_on

    @property
    def screen_size(self):
        """Get screen size."""
        return (self.width * self.scale, self.height * self.scale)

    def render(self):
        """Render chip8 sprites."""
        self.screen.fill(self.color_off)
        for y, row in enumerate(self.chip8.display):
            x = 0
            bits = row
            while bits:
                cell = 0x8000000000000000 & bits
                if cell:
                    draw_pixel(x, y, self.scale, self.color_on)
                bits &= 0x7fffffffffffffff
                bits <<= 1
                x += 1
        pygame.display.update()

    def init_screen(self):
        """Initialize screen."""
        pygame.init()
        pygame.display.set_caption("Chippy")
        self.screen = pygame.display.set_mode(self.screen_size)

    def handle_key_event_when_running(self, event):
        """Handle KEYUP and KEYDOWN events in RUN mode."""
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SEMICOLON:
                self.chip8.status = Mode.PAUSE
                return
            key = convert_key(event.key)
            if key is not None:
                press(self.chip8, key)
        elif event.type == pygame.KEYUP:
            key = convert_key(event.key)
            if key is not None:
                release(self.chip8, key)

    def handle_key_event_when_paused(self, event):
        """Handle KEYUP and KEYDOWN events when paused."""
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                self.chip8.status = Mode.RUN
                return

    def handle_events(self):
        """Handle Pygame events and Chippy interrupts.."""
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.chip8.status = Mode.STOP
                return
            if self.chip8.status == Mode.RUN:
                self.handle_key_event_when_running(event)
            elif self.chip8.status == Mode.PAUSE:
                self.handle_key_event_when_paused(event)



import random

class ExecutionUnit:
    def __init__(self, chip8):
        self.vm = chip8

    def op_0nnn(self, nnn):
        """Jump to routine at nnn."""
        raise NotImplementedError

    def op_00e0(self):
        """Clear chip8 display."""
        self.vm.initialize_display()

    def op_00ee(self):
        """Return from subroutine."""
        self.vm.stack_pointer -= 1
        self.vm.jump(self.vm.stack[self.vm.stack_pointer])

    def op_1nnn(self, nnn):
        """Jump to location nnn."""
        self.vm.jump(nnn)

    def op_2nnn(self, nnn):
        """Call subroutine at nnn."""
        self.vm.stack[self.vm.stack_pointer] = self.vm.program_counter
        self.vm.stack_pointer += 1
        self.vm.jump(nnn)

    def op_3xkk(self, x, kk):
        """Skip the next instruction if Vx == kk."""
        if self.vm.registers[x] == kk:
            self.vm.increment()

    def op_4xkk(self, x, kk):
        """Skip next instruction if Vx != kk."""
        if self.vm.registers[x] != kk:
            self.vm.increment()

    def op_5xy0(self, x, y):
        """Skip next instruction if Vx = Vy."""
        if self.vm.registers[x] == self.vm.registers[y]:
            self.vm.increment()

    def op_6xkk(self, x, kk):
        """Put kk into Vx."""
        self.vm.registers[x] = kk

    def op_7xkk(self, x, kk):
        """Add kk to Vx."""
        self.vm.registers[x] = (self.vm.registers[x] + kk) & 0xff

    def op_8xy0(self, x, y):
        """Set Vx = Vy."""
        self.vm.registers[x] = self.vm.registers[y]

    def op_8xy1(self, x, y):
        """Set Vx = Vx OR Vy."""
        self.vm.registers[x] |= self.vm.registers[y]

    def op_8xy2(self, x, y):
        """Set Vx = Vx AND Vy."""
        self.vm.registers[x] &= self.vm.registers[y]

    def op_8xy3(self, x, y):
        """Set Vx = Vx XOR Vy."""
        self.vm.registers[x] ^= self.vm.registers[y]

    def op_8xy4(self, x, y):
        """Add Vy to Vx and set Vf to the carry bit."""
        assert x != 0xf
        total = self.vm.registers[x] + self.vm.registers[y]
        self.vm.registers[x] = total & 0xff
        self.vm.registers[0xf] = int(total > 0xff)

    def op_8xy5(self, x, y):
        """Subtract Vy from Vx and set Vf to 1 if there's no borrow."""
        assert x != 0xf
        difference = self.vm.registers[x] - self.vm.registers[y]
        self.vm.registers[x] = difference & 0xff
        self.vm.registers[0xf] = int(difference > 0)   # or >= 0? (see Cowgod reference)

    def op_8xy6(self, x, y):
        """Set Vx = Vx >> 1, and set Vf to the LSB prior to the shift.

        NOTE Descriptions differ between
        - http://mattmik.com/files/chip8/mastering/chip8.html
        - http://devernay.free.fr/hacks/chip8/C8TECH10.HTM

        Vx should be shifted, not Vy.
        Vy should be ignored.
        """
        assert x != 0xf
        y = x
        self.vm.registers[0xf] = self.vm.registers[y] & 0x01
        self.vm.registers[x] = self.vm.registers[y] >> 1

    def op_8xy7(self, x, y):
        """Set Vx = Vy - Vx and set Vf to 1 if there's no borrow."""
        assert x != 0xf
        difference = self.vm.registers[y] - self.vm.registers[x]
        self.vm.registers[x] = difference & 0xff
        self.vm.registers[0xf] = int(difference > 0)   # or >= 0?

    def op_8xye(self, x, y):
        """Set Vx = Vy << 1, and set Vf to the MSB prior to the shift.

        NOTE Descriptions differ between
        - http://mattmik.com/files/chip8/mastering/chip8.html
        - http://devernay.free.fr/hacks/chip8/C8TECH10.HTM
        """
        # NOTE should Vx or Vy be shifted? (see op_8xy6)
        assert x != 0xf
        y = x
        self.vm.registers[0xf] = (self.vm.registers[y] >> 7) & 0x01
        self.vm.registers[x] = (self.vm.registers[y] << 1) & 0xff

    def op_9xy0(self, x, y):
        """Skip next instruction if Vx != Vy."""
        if self.vm.registers[x] != self.vm.registers[y]:
            self.vm.increment()

    def op_annn(self, nnn):
        """Set I to nnn."""
        self.vm.I = nnn

    def op_bnnn(self, nnn):
        """Jump to nnn + V0."""
        self.vm.jump((nnn + self.vm.registers[0]) & 0xfff)

    def op_cxkk(self, x, kk):
        """Set Vx = random byte & kk."""
        self.vm.registers[x] = random.randint(0x00, 0xff) & kk

    def op_dxyn(self, x, y, nibble):
        """Display n-byte sprite starting at memory location I at (Vx, Vy).

        The sprite is 8 pixels wide and n pixels tall.
        Set Vf = 1 iff any set pixels are unset.
        The sprite is drawn by XORing it with the display.

        Out of screen parts of sprites wrap around to the other side.
        """
        sprite = self.vm.ram[self.vm.I:self.vm.I + nibble]
        X = self.vm.registers[x] & 0x3f
        Y = self.vm.registers[y]
        self.vm.registers[0xf] = 0

        shift_amount = 56 - X
        for i, row in enumerate(sprite):
            if shift_amount < 0:
                shifted_row = row >> -shift_amount
                offscreen = row << (64 + shift_amount)
                offscreen &= 0xffffffffffffffff
                shifted_row |= offscreen
            else:
                shifted_row = row << shift_amount

            Y32 = (Y + i) & 0x1f
            xor = self.vm.display[Y32] ^ shifted_row
            unset = self.vm.display[Y32] & shifted_row
            self.vm.display[Y32] = xor

            if shift_amount < 0:
                offscreen = unset >> (64 + shift_amount)
                unset <<= -shift_amount
                unset |= offscreen
            else:
                unset >>= shift_amount
            unset &= 0xff
            self.vm.registers[0xf] |= unset
        self.vm.registers[0xf] = 1 if self.vm.registers[0xf] else 0

    def op_ex9e(self, x):
        """Skip next instruction if key with the value of Vx is pressed."""
        pressed = (self.vm.keypad >> (self.vm.registers[x] & 0xf)) & 0x1
        if pressed:
            self.vm.increment()

    def op_exa1(self, x):
        """Skip next instruction if key with the value of Vx is not pressed."""
        pressed = (self.vm.keypad >> (self.vm.registers[x] & 0xf)) & 0x1
        if not pressed:
            self.vm.increment()

    def op_fx07(self, x):
        """Set Vx to the delay timer value."""
        self.vm.registers[x] = self.vm.delay_timer

    def op_fx0a(self, x):
        """Wait for a key press and store the value of the key in Vx."""
        self.vm.waiting.append(x)

    def op_fx15(self, x):
        """Set the delay timer to Vx."""
        self.vm.delay_timer = self.vm.registers[x]

    def op_fx18(self, x):
        """Set the sound timer to Vx."""
        self.vm.sound_timer = self.vm.registers[x]

    def op_fx1e(self, x):
        """Add Vx to I."""
        self.vm.I = (self.vm.I + self.vm.registers[x]) & 0xffff

    def op_fx29(self, x):
        """Set I to location of sprite for digit in Vx."""
        self.vm.I = (self.vm.registers[x] & 0x0f) * 5

    def op_fx33(self, x):
        """Store the BCD representation of Vx in memory locations I, I+1 and I+2."""
        b = self.vm.registers[x] // 100
        c = (self.vm.registers[x] // 10) % 10
        d = self.vm.registers[x] % 10
        self.vm.ram[self.vm.I] = b
        self.vm.ram[self.vm.I+1] = c
        self.vm.ram[self.vm.I+2] = d

    def op_fx55(self, x):
        """Store registers V0 to Vx (inclusive) in memory starting at location I.

        The value of I doesn't get incremented, unlike what the ff. guide says.
        - http://mattmik.com/files/chip8/mastering/chip8.html
        """
        self.vm.ram[self.vm.I:self.vm.I + x+1] = self.vm.registers[:x+1]

    def op_fx65(self, x):
        """Read registers V0 through Vx (inclusive) from memory starting at I.

        The value of I doesn't get incremented, unlike what the ff. guide says.
        - http://mattmik.com/files/chip8/mastering/chip8.html
        """
        self.vm.registers[:x+1] = self.vm.ram[self.vm.I:self.vm.I + x + 1]



import setuptools

with open("README.md") as fp:
    long_description = fp.read()

setuptools.setup(
    name="chippy-emu",
    version="0.1.4",
    author="Levi Gruspe",
    author_email="mail.levig@gmail.com",
    description="Chip-8 interpreter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lggruspe/chippy",
    packages=setuptools.find_packages(),
    package_data={
        "chippy": ["roms/*", "data/*.wav"],
    },
    classifiers=[
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Games/Entertainment",
    ],
    install_requires=["pygame"],
    python_requires=">=3.7",
)
